import numpy as np
import scipy.sparse.linalg as sla
import scipy.sparse as sp

try:
    import torch
    from torch.autograd import Variable
except:
    pass


import re
from abc import ABCMeta, abstractmethod


def block(rows, dtype=None, arrtype=None):
    if (not _is_list_or_tup(rows)) or len(rows) == 0 or \
       np.any([not _is_list_or_tup(row) for row in rows]):
        raise RuntimeError('''
Unexpected input: Expected a non-empty list of lists.
If you are interested in helping expand the functionality
for your use case please send in an issue or PR at
http://github.com/bamos/block''')

    rowLens = [len(row) for row in rows]
    if len(np.unique(rowLens)) > 1:
        raise RuntimeError('''
Unexpected input: Rows are not the same length.
Row lengths: {}'''.format(rowLens))

    nRows = len(rows)
    nCols = rowLens[0]
    rowSizes = np.zeros(nRows, dtype=int)
    colSizes = np.zeros(nCols, dtype=int)

    backend = _get_backend(rows, dtype, arrtype)

    for i, row in enumerate(rows):
        for j, elem in enumerate(row):
            if backend.is_complete(elem):
                rowSz, colSz = backend.extract_shape(elem)
                rowSizes[i] = rowSz
                colSizes[j] = colSz
            elif hasattr(elem, 'shape'):
                rowSz, colSz = elem.shape
                rowSizes[i] = rowSz
                colSizes[j] = colSz
            elif hasattr(elem, 'size'):
                rowSz, colSz = elem.size()
                rowSizes[i] = rowSz
                colSizes[j] = colSz

    cRows = []
    for row, rowSz in zip(rows, rowSizes):
        rowSz = int(rowSz)
        if rowSz == 0:
            continue
        cCol = []
        for elem, colSz in zip(row, colSizes):
            colSz = int(colSz)
            if colSz == 0:
                continue
            # TODO: Check types.
            if backend.is_complete(elem):
                cElem = elem
            elif isinstance(elem, float) or isinstance(elem, int):
                cElem = backend.build_full((rowSz, colSz), elem)
            elif isinstance(elem, str):
                if elem == 'I':
                    assert(rowSz == colSz)
                    cElem = backend.build_eye(rowSz)
                elif elem == '-I':
                    assert(rowSz == colSz)
                    cElem = -backend.build_eye(rowSz)
                else:
                    assert(False)
            else:
                cElem = backend.convert(elem)
            cCol.append(cElem)
        cRows.append(cCol)

    return backend.build(cRows)


def block_diag(elems, dtype=None, arrtype=None):
    n = len(elems)
    return block([[0] * i + [elem] + [0] * (n - 1 - i)
                  for i, elem in enumerate(elems)],
                 dtype=dtype, arrtype=arrtype)


def block_tridiag(main, upper, lower):
    n = len(main)
    assert len(main) == len(upper) + 1
    assert len(main) == len(lower) + 1
    mat = ()
    for i in range(n):
        tup = ()
        for j in range(n):
            if (i==j):   tup = (*tup, main[i])
            elif (i==j-1): tup = (*tup, upper[-i])
            elif (i==j+1): tup = (*tup, lower[i-1])
            else: tup = (*tup,0)
        mat = (*mat,tup)
    return block(mat)

def _is_list_or_tup(x):
    return isinstance(x, list) or isinstance(x, tuple)


def _get_backend(rows, dtype, arrtype):
    if arrtype == np.ndarray and dtype is not None:
        return NumpyBackend(arrtype, dtype)
    elif arrtype == sla.LinearOperator:
        return LinearOperatorBackend(dtype)
    elif arrtype is not None and re.search('torch\..*Tensor', repr(arrtype)):
        return TorchBackend(dtype)
    elif arrtype is not None and re.search('torch\..*(Variable|Parameter)', repr(arrtype)):
        return TorchVariableBackend(dtype)
    else:
        npb = NumpyBackend()
        tb = TorchBackend()
        lob = LinearOperatorBackend()
        tvb = TorchVariableBackend()
        for row in rows:
            for elem in row:
                if npb.is_complete(elem) and elem.size > 0:
                    if dtype is None:
                        dtype = type(elem[0, 0])
                    if arrtype is None:
                        arrtype = type(elem)
                    return NumpyBackend(dtype, arrtype)
                elif tb.is_complete(elem):
                    return TorchBackend(type(elem))
                elif lob.is_complete(elem):
                    return LinearOperatorBackend(elem.dtype)
                elif tvb.is_complete(elem):
                    return TorchVariableBackend(type(elem.data))

    assert(False)


class Backend():
    __metaclass__ = ABCMeta

    @abstractmethod
    def extract_shape(self, x): pass

    @abstractmethod
    def build_eye(self, n): pass

    @abstractmethod
    def build_full(self, shape, fill_val): pass

    @abstractmethod
    def convert(self, x): pass

    @abstractmethod
    def build(self, rows): pass

    @abstractmethod
    def is_complete(self, rows): pass


class NumpyBackend(Backend):

    def __init__(self, dtype=None, arrtype=None):
        self.dtype = dtype
        self.arrtype = arrtype

    def extract_shape(self, x):
        return x.shape

    def build_eye(self, n):
        return np.eye(n)

    def build_full(self, shape, fill_val):
        return np.full(shape, fill_val, self.dtype)

    def convert(self, x):
        assert(False)

    def build(self, rows):
        return np.bmat(rows)

    def is_complete(self, x):
        return isinstance(x, np.ndarray)


class TorchBackend(Backend):

    def __init__(self, dtype=None):
        self.dtype = dtype

    def extract_shape(self, x):
        return x.size()

    def build_eye(self, n):
        return torch.eye(n).type(self.dtype)

    def build_full(self, shape, fill_val):
        return fill_val * torch.ones(*shape).type(self.dtype)

    def convert(self, x):
        assert(False)

    def build(self, rows):
        compRows = []
        for row in rows:
            compRows.append(torch.cat(row, 1))
        return torch.cat(compRows)

    def is_complete(self, x):
        return (re.search('torch\..*Tensor', str(x.__class__)) is not None) \
            and x.ndimension() == 2


class TorchVariableBackend(TorchBackend):
    def build_eye(self, n):
        return Variable(super().build_eye(n))

    def build_full(self, shape, fill_val):
        return Variable(super().build_full(shape, fill_val))

    def convert(self, x):
        if TorchBackend.is_complete(self, x):
            return Variable(x)
        assert(False)

    def is_complete(self, x):
        return re.search('torch\..*(Variable|Parameter)', str(x.__class__))


class LinearOperatorBackend(Backend):
    def __init__(self, dtype=None):
        self.dtype = dtype

    def extract_shape(self, x):
        return x.shape

    def build_eye(self, n):
        def identity(v): return v
        return sla.LinearOperator(shape=(n, n),
                                  matvec=identity,
                                  rmatvec=identity,
                                  matmat=identity,
                                  dtype=self.dtype)

    def build_full(self, shape, fill_val):
        m, n = shape
        if fill_val == 0:
            return shape
        else:
            def matvec(v):
                return v.sum() * fill_val * np.ones(m)

            def rmatvec(v):
                return v.sum() * fill_val * np.ones(n)

            def matmat(M):
                return M.sum(axis=0) * fill_val * np.ones((m, M.shape[1]))

            return sla.LinearOperator(shape=shape,
                                      matvec=matvec,
                                      rmatvec=rmatvec,
                                      matmat=matmat,
                                      dtype=self.dtype)

    def convert(self, x):
        if (isinstance(x, (np.ndarray, sp.spmatrix))):
            return sla.aslinearoperator(x)
        else:
            assert(False)

    def build(self, rows):
        col_sizes = [lo.shape[1] if self.is_complete(lo) else lo[1]
                     for lo in rows[0]]
        col_idxs = np.cumsum([0] + col_sizes)
        row_sizes = [row[0].shape[0] if self.is_complete(row[0]) else row[0][0]
                     for row in rows]
        row_idxs = np.cumsum([0] + row_sizes)
        m, n = sum(row_sizes), sum(col_sizes)

        def matvec(v):
            out = np.zeros(m)
            for row, i, j in zip(rows, row_idxs[:-1], row_idxs[1:]):
                out[i:j] = sum(lo.matvec(v[k:l]) for lo, k, l in
                               zip(row, col_idxs[:-1], col_idxs[1:])
                               if self.is_complete(lo))
            return out

        # The transposed list
        cols = zip(*rows)

        def rmatvec(v):
            out = np.zeros(n)
            for col, i, j in zip(cols, col_idxs[:-1], col_idxs[1:]):
                out[i:j] = sum(lo.rmatvec(v[k:l]) for lo, k, l in
                               zip(col, row_idxs[:-1], row_idxs[1:])
                               if self.is_complete(lo))
            return out

        def matmat(M):
            out = np.zeros((m, M.shape[1]))
            for row, i, j in zip(rows, row_idxs[:-1], row_idxs[1:]):
                out[i:j] = sum(lo.matmat(M[k:l]) for lo, k, l in
                               zip(row, col_idxs[:-1], col_idxs[1:])
                               if self.is_complete(lo))
            return out

        return sla.LinearOperator(shape=(m, n),
                                  matvec=matvec,
                                  rmatvec=rmatvec,
                                  matmat=matmat,
                                  dtype=self.dtype)

    def is_complete(self, x):
        return isinstance(x, sla.LinearOperator)


def get_toeplitz(rows, cols=None):
    # https://stackoverflow.com/questions/36464191/toeplitz-matrix-of-toeplitz-matrix
    '''This function constructs the toepliz matrix from the
    input block rows. If cols is None, the matrix is Hermitian.'''
    if cols is None:
        cols  = rows[1:][::-1]
    # add (0,0) bloch diagonal block at front
    if len(cols) == len(rows) - 1:
        cols = np.insert(cols, 0, rows[0], axis=0)
    # cols contains (0,0) bloch diagonal block
    assert len(cols) == len(rows)
    dtype = rows[0].dtype
    m, n  = rows[0].shape
    n_r   = len(rows)
    mat = np.zeros((n_r, m, n_r, n), dtype=dtype)
    rows_it  = iter(rows)
    cols_it  = iter(cols)
    for j, entry in enumerate(rows_it):
        col_mat = next(cols_it)
        for i in range(n_r-j):
            mat[i,:,i+j,:] = entry
            mat[j+i,:,i,:] = col_mat
    mat.shape = (n_r*m, n_r*n)
    return mat

def test_toepliz():
    ## test
    rows = [np.arange(4).reshape(2,2),
            np.arange(4,8).reshape(2,2),
            np.arange(8,12).reshape(2,2)]
    cols = [np.arange(12,16).reshape(2,2),
            np.arange(16,20).reshape(2,2)]
    mat  = get_toeplitz(np.array(rows), np.array(cols))

    assert np.allclose(mat, np.array([[ 0,  1,  4,  5,  8,  9],
                                   [ 2,  3,  6,  7, 10, 11],
                                   [12, 13,  0,  1,  4,  5],
                                   [14, 15,  2,  3,  6,  7],
                                   [16, 17, 12, 13,  0,  1],
                                   [18, 19, 14, 15,  2,  3]]))



import numpy as np
from scipy.sparse import coo_matrix
from ase.units import Hartree

import re
from pathlib import Path

from ..misc import tri2full


float_match = '([-+]?\d*\.\d+)'
int_match = '(\d+)'

def read_complex_bincsr(filename, dtype=np.float64):
    """Read complex matrix stored in .csr CP2K format.
    
    NOTE :: The returned matrix is in units of [eV].

    Args:
        filename : (str, file object)
        dtype : (np.dtype)
            data type of real and immaginary parts.
    """
    dt = np.dtype([
        ('o1', np.uint32),
        ('x', np.uint32),
        ('y', np.uint32),
        ('real', dtype),
        ('imag', dtype),
        ('o2', np.uint32)])
    
    coo = np.fromfile(filename, dt, offset=0)
    data = coo['real'] + 1.j * coo['imag']
    mat = coo_matrix((data,(coo['x']-1,coo['y']-1))).toarray()
    mat *= Hartree
    tri2full(mat,UL='U')
    return mat


def read_real_bincsr(filename, dtype=np.float64):
    """Read real matrix stored in .csr CP2K format.
    
    NOTE :: The returned matrix is in units of [eV].

    Args:
        filename : (str, file object)
        dtype : (np.dtype)
            data type of real and immaginary parts.
    """
    dt = np.dtype([
        ('o1', np.uint32),
        ('x', np.uint32),
        ('y', np.uint32),
        ('data', dtype),
        ('o2', np.uint32)])
    
    coo = np.fromfile(filename, dt, offset=0)
    mat = coo_matrix((coo['data'],(coo['x']-1,coo['y']-1))).toarray()
    mat *= Hartree
    tri2full(mat,UL='U')
    return mat


def read_fermi(filename):
    """Read Fermi level (in [eV]) from output file."""
    pattern = 'Fermi energy:' + '\s+' + float_match
    matches = re.findall(pattern, open(filename,'r').read())
    fermi = np.unique(np.array(matches, dtype=float))
    assert fermi.size==1, "Found multiple Fermi levels that don't match! {fermi*Hartree}"
    return fermi[0]*Hartree


def read_mp_kmesh(filename):
    """Read Monkhorst-Pack k-point grid from output file."""
    pattern = 'K-Point grid' + 3 * ('\s+' + int_match)
    match = re.search(pattern, open(filename,'r').read())
    size = [int(match.group(i)) for i in range(1,4)]
    return size


def read_kpts(filename):
    """Read list of k-points from output file."""
    fp = open(filename,'r').read()
    pattern = 'List of Kpoints \[2 Pi/Bohr\]' + '\s+' + int_match
    num_kpts = int(re.search(pattern, fp).group(1))
    kpts = np.empty((num_kpts,3))
    weights = np.empty(num_kpts)
    pattern = 'BRILLOUIN\|.+?' + 4 * ('\s+' + float_match)
    for ik, kpt_info in enumerate(re.finditer(pattern, fp)):
        weights[ik] = float(kpt_info.group(1))
        kpts[ik] = [float(kpt) for kpt in kpt_info.group(2,3,4)]
        if ik == num_kpts-1: break
    return kpts, weights


def read_basis_dict(filename):
    """Read basis function info as dictionary."""
    basis = {}
    fp = open(filename,'r').read()
    # for kind in re.finditer('ATOMIC KIND INFORMATION', fp):
    #     symbol = re.search('Atomic kind:\s+(\w)', fp).group(1)
    for kind in re.finditer('Atomic kind:\s+(\w+)', fp):
        symbol = kind.group(1)
        nao = re.search('Number of spherical basis functions:\s+(\d+)', fp).group(1)
        basis[symbol] = int(nao)
    return basis


def read_reversed_from_folder(path, project='*', parser=read_complex_bincsr):
    """Read Hamilton and Overlap matrices in reverse order from folder.
    
    Both CP2K and PYQT apply inversion symmetry along the x-direction.
    However:
    e.g. with Monkhorst-Pack mesh = (2,2,2)
    PYQT considers the second half of the BZ:
       [ 0.25, -0.25, -0.25],
       [ 0.25, -0.25,  0.25],
       [ 0.25,  0.25, -0.25],
       [ 0.25,  0.25,  0.25]]
    CP2K considers the first half of the BZ:
       [ -0.25, -0.25, -0.25],
       [ -0.25, -0.25,  0.25],
       [ -0.25,  0.25, -0.25],
       [ -0.25,  0.25,  0.25]]

    Hence we 
        (i) read the k-point matrices from last k-point to first k-point
        (ii) transpose each matrix since A[k] = A[-k].T


    Args:
        path : (str, pathlib.Path object)
            path to folder
        project : (str), optional
            name of the project. 
            default is read all Hamilton ans Overlap in folder.
        parser : (callable), optional
            used to read each matrix file. 
            default is read complex matrix in binary format.
    """
    path = Path(path)
    h_cc_k_files = list(path.glob(f"{project}-KS_SPIN_1_K_*-1_0.csr"))
    s_cc_k_files = list(path.glob(f"{project}-S_SPIN_1_K_*-1_0.csr"))

    pattern = '.+?_K_(\d+)-.+?'
    h_cc_k_files = sorted(h_cc_k_files, key=lambda p: int(re.search(pattern, p.stem).group(1)), reverse=True)
    s_cc_k_files = sorted(s_cc_k_files, key=lambda p: int(re.search(pattern, p.stem).group(1)), reverse=True)
    
    read = lambda file: parser(file).T
    h_cc_k = np.stack(list(map(read, h_cc_k_files)))
    s_cc_k = np.stack(list(map(read, s_cc_k_files)))
    return h_cc_k, s_cc_k


#Cutoff for cutting block matrices
from qtpyt import xp, xla

solve = xp.linalg.solve #la.solve
inv = xp.linalg.inv #la.inv
dot = xp.dot
dagger = lambda mat: xp.conj(mat.T)


def coupling_method_1N(m_qii, m_qij, m_qji):
    """
    TODO :: 
    (1) a_mm = solve(..,m_qij) inplace m_qij
    """
    N = len(m_qii)
    
    q = 1
    a_mm = solve(m_qii[q-1], m_qij[q-1])  
    m_qii[q] -= dot(m_qji[q-1], a_mm)
    c_mm = a_mm
    for q in range(2,N):
        a_mm = solve(m_qii[q-1], m_qij[q-1])  
        m_qii[q] -= dot(m_qji[q-1], a_mm)
        c_mm = dot(c_mm, a_mm)

    c_mm *= (-1)**(N-1)
    # solve xA=B =: x=BA^-1 
    # (xA)^T = B^T 
    # A^Tx^T = B^T
    g_1N = solve(m_qii[N-1].T,c_mm.T).T

    return g_1N


def coupling_method_N1(m_qii, m_qij, m_qji):
    """
    TODO :: 
    (1) a_mm = solve(..,m_qji) inplace m_qji
    """
    N = len(m_qii)
    
    q = N-2
    a_mm = solve(m_qii[q+1], m_qji[q])
    m_qii[q] -= dot(m_qij[q], a_mm)
    c_mm = a_mm
    for q in range(N-3,-1,-1):
        a_mm = solve(m_qii[q+1], m_qji[q])
        m_qii[q] -= dot(m_qij[q], a_mm)
        c_mm = dot(c_mm, a_mm)

    c_mm *= (-1)**(N-1)
    # solve xA=B =: x=BA^-1 
    # (xA)^T = B^T 
    # A^Tx^T = B^T
    g_N1 = solve(m_qii[0].T,c_mm.T).T

    return g_N1


def spectral_method(m_qii, m_qij, m_qji, gamma_L, gamma_R):

    N = len(m_qii)
    
    n_qii = [None for _ in range(N)]
    n_qji = [None for _ in range(N-1)]
    n_qij = [None for _ in range(N-1)]

    n_qii[-1] = m_qii[-1]
    for q in range(N-2,-1,-1):
        n_qji[q] = solve(n_qii[q+1], m_qji[q]) # can't overwrite
        n_qii[q] = m_qii[q] - dot(m_qij[q], n_qji[q]) # can't overwrite

    for q in range(1,N):
        m_qij[q-1] = solve(m_qii[q-1], m_qij[q-1]) # can overwrite
        m_qii[q] -= dot(m_qji[q-1], m_qij[q-1]) # can overwrite

    a_mm = solve(n_qii[0], gamma_L)
    n_qii[0] = solve(n_qii[0].conj(), a_mm.T).T
    for q in range(N-1):
        a_mm = -n_qji[q]
        n_qji[q] = dot(a_mm, n_qii[q], out=n_qji[q])
        n_qij[q] = dot(n_qii[q], dagger(a_mm))#, out=n_qij[q])
        n_qii[q+1] = dot(a_mm, n_qij[q])#, out=n_qii[q+1])

    a_mm = solve(m_qii[-1], gamma_R)
    m_qii[-1] = solve(m_qii[-1].conj(), a_mm.T).T
    for q in range(N-2,-1,-1):
        a_mm = -m_qij[q]
        m_qij[q] = dot(a_mm, m_qii[q+1], out=m_qij[q])
        m_qji[q] = dot(m_qii[q+1], dagger(a_mm))#, out=m_qji[q].T)
        m_qii[q] = dot(a_mm, m_qji[q], out=m_qii[q])

    return (n_qii, n_qij, n_qji), (m_qii, m_qij, m_qji)


def dyson_method(m_qii, m_qij, m_qji, trans=False):

    N = len(m_qii)

    # Left connected green's function
    grL_qii = m_qii#[None for _ in range(N)]
    # Initalize
    grL_qii[0] = inv(m_qii[0]) # ([eS-H]_11-Sigma_L)^-1

    # Left connected recursion
    for q in range(1, N):
        a_ij = dot(grL_qii[q-1], m_qij[q-1])
        grL_qii[q] = inv(m_qii[q] - dot(m_qji[q-1], a_ij))

    if trans:
        # First row green's function
        gr_1N = grL_qii[0]
        for q in range(1, N):
            gr_1N = dot(gr_1N, dot(m_qij[q-1],grL_qii[q]))
        gr_1N *= (-1)**(N-1)

    # Full green's function
    Gr_qii = [None for _ in range(N)]
    Gr_qji = [None for _ in range(N-1)]
    Gr_qij = [None for _ in range(N-1)]
    # Initialize
    Gr_qii[-1] = grL_qii[-1] # G_NN = gL_NN = ([eS-H]_NN - [eS-H]_NN-1 * grL_N-1N-1 * [eS-H]_N-1N - Sigma_R)^-1

    # Full recursion
    for q in range(N-2, -1, -1):
        a_ij = dot(-grL_qii[q], m_qij[q])
        a_ji = dot(-m_qji[q], grL_qii[q])
        Gr_qji[q] = dot(Gr_qii[q + 1], a_ji)
        Gr_qij[q] = dot(a_ij, Gr_qii[q + 1])
        Gr_qii[q] = grL_qii[q] + dot(a_ij, Gr_qji[q])

    if trans:
        return gr_1N, (Gr_qii, Gr_qij, Gr_qji)
    return (Gr_qii, Gr_qij, Gr_qji)

    # gnL_qii = [None for _ in range(N)]
    # s_in = [None for _ in range(N)]
    # s_in[0] = gamma_L; s_in[-1] = gamma_R

    # gnL_qii[q] = dot(grL_qii[q], dot(s_in[0], grL_qii[q].conj()))    
    # for q in range(1, N):
    #     sl = dot(m_qji[q-1], dot(gnL_qii[q-1], m_qij[q-1].conj()))
    #     if s_in[q] is not None: sl += s_in[q]
    #     gnL_qii[q] = dot(grL_qii[q], dot(sl, grL_qii[q].conj()))
    
    # Gn_qii = [None for _ in range(N)]
    # Gn_qij = [None for _ in range(N-1)]
    # Gn_qji = [None for _ in range(N-1)]

    # Gn_qii[-1] = gnL_qii[-1]
    # for q in range(N-2,-1,-1):
    #     a = dot(Gn_qii[q+1], dot(m_qji[q].conj(), grL_qii[q].conj()))
    #     Gn_qji[q] = - (dot(Gr_qii[q+1], dot(m_qji[q], gnL_qii[q])) + a)
    #     Gn_qii[q] = gnL_qii[q] + \
    #                 dot(grL_qii[q], dot(m_qij[q],  a)) - \
    #                 dot(gnL_qii[q], dot(m_qij[q].conj(), Gr_qji[q].conj())) - \
    #                 dot(Gr_qij[q], dot(m_qji[q], gnL_qii[q]))
    #     Gn_qij[q] = dagger(Gn_qji[q])

    # return (Gn_qii, Gn_qij, Gn_qji)


# def overlap_method(m_qii, m_qij, m_qji):

#     N = len(m_qii)
#     k = xp.argmin([m_qii[q].shape[0] for q in range(N)])

#     for q in range(1,k):
#         a_mm = solve(m_qii[q-1], m_qij[q-1]) # Here we could overwrite m_qij[q-1]  
#         m_qii[q] -= dot(m_qji[q-1], a_mm)

#     q = k
#     sL_kk = dot(m_qji[q-1], solve(m_qii[q-1], m_qij[q-1]))  # Here we could overwrite m_qji[q+1]

#     for q in range(N-2,k,-1):
#         a_mm = solve(m_qii[q+1], m_qji[q])
#         m_qii[q] -= dot(m_qij[q], a_mm)
    
#     q = k
#     sR_kk = dot(m_qij[q], solve(m_qii[q+1], m_qji[q]))

#     gammaL_kk = 1.j * (sL_kk - sL_kk.T.conj())
#     gammaR_kk = 1.j * (sR_kk - sR_kk.T.conj())
#     m_qii[k] -= sL_kk + sR_kk
#     g_kk = inv(m_qii[k]) # Assert allclose with G[kk].inv() OK

#     a_mm = solve(m_qii[k], gammaL_kk)
#     b_mm = solve(dagger(m_qii[k]), gammaR_kk)

#     return trace(dot(a_mm, b_mm)).real


import numpy as np
from .tools import subdiagonalize, rotate_matrix, dagger, \
                   order_diagonal, cutcoupling, get_subspace


from .tk_gpaw import subdiagonalize_atoms, get_bfs_indices, \
                     extract_orthogonal_subspaces, flatten

class CoupledHamiltonian:

    def __init__(self, H, S, selfenergies):
        self.H = H
        self.S = S
        self.selfenergies = selfenergies

    def align_bf(self, align_bf):
        h_mm = self.H
        s_mm = self.S
        h1_ii = self.selfenergies[0].h_ii
        if align_bf is not None:
            diff = ((h_mm[align_bf, align_bf] - np.real(h1_ii[align_bf, align_bf])) /
                    s_mm[align_bf, align_bf])
            # print('# Aligning scat. H to left lead H. diff=', diff)
            h_mm -= diff * s_mm
        return diff

    def apply_rotation(self, c_mm):
          h_mm = self.H
          s_mm = self.S
          h_mm[:] = rotate_matrix(h_mm, c_mm)
          s_mm[:] = rotate_matrix(s_mm, c_mm)
          # Rotate coupling between lead and central region
          for alpha, sigma in enumerate(self.selfenergies):
              sigma.h_im[:] = np.dot(sigma.h_im, c_mm)
              sigma.s_im[:] = np.dot(sigma.s_im, c_mm)

    def diagonalize(self, apply=False):
        nbf = len(self.H)
        return self.subdiagonalize_bfs(range(nbf), apply)

    def subdiagonalize_bfs(self, bfs, apply=False):
      bfs = np.array(bfs)
      h_mm = self.H
      s_mm = self.S
      ht_mm, st_mm, c_mm, e_m = subdiagonalize(h_mm, s_mm, bfs)
      if apply:
          self.apply_rotation(c_mm)
          return

      c_mm = np.take(c_mm, bfs, axis=0)
      c_mm = np.take(c_mm, bfs, axis=1)
      return ht_mm, st_mm, e_m.real, c_mm

    def cutcoupling_bfs(self, bfs, apply=False):
      bfs = np.array(bfs)
      h_pp = self.H.copy()
      s_pp = self.S.copy()
      cutcoupling(h_pp, s_pp, bfs)
      if apply:
          self.H = h_pp
          self.S = s_pp
          for alpha, sigma in enumerate(self.selfenergies):
              for m in bfs:
                  sigma.h_im[:, m] = 0.0
                  sigma.s_im[:, m] = 0.0
      return h_pp, s_pp

    def take_bfs(self, bfs, apply=False):
        nbf = len(self.H)
        c_mm = np.eye(nbf).take(bfs,1)
        h_pp = rotate_matrix(self.H, c_mm)
        s_pp = rotate_matrix(self.S, c_mm)
        if apply:
            self.H = h_pp
            self.S = s_pp
            for alpha, sigma in enumerate(self.selfenergies):
                sigma.h_im = np.dot(sigma.h_im, c_mm)
                sigma.s_im = np.dot(sigma.s_im, c_mm)
                sigma.sigma_mm = np.empty(self.H.shape, complex)
        return h_pp, s_pp, c_mm

    def lowdin_rotation(self, apply=False):
      h_mm = self.H
      s_mm = self.S
      eig, rot_mm = np.linalg.eigh(s_mm)
      eig = np.abs(eig)
      rot_mm = np.dot(rot_mm / np.sqrt(eig), dagger(rot_mm))
      if apply:
          self.apply_rotation(rot_mm)
          return

      return rot_mm

    def order_diagonal(self, apply=False):
      h_mm = self.H
      s_mm = self.S
      ht_mm, st_mm, c_mm = order_diagonal(h_mm, s_mm)
      if apply:
          self.apply_rotation(c_mm)
          return

      return ht_mm, st_mm, c_mm

    def subdiagonalize_atoms(self, calc, a=None, apply=False):
        h_mm = self.H
        s_mm = self.S
        c_MM, e_aj = subdiagonalize_atoms(calc, h_mm, s_mm, a)
        if apply:
            #Apply subdiagonalization
            self.apply_rotation(c_MM)
            return
        return c_MM, e_aj


    # def take_bfs_activespace(self, calc, a, key=lambda x: abs(x)<np.inf,
    #                          cutoff=np.inf, include=False, orthogonal=True):
    #     '''
    #         key     := Apply condition to eigenvalues of subdiagonalized atoms.
    #         cutoff  := Cutoff for effective embedding to incudle in active space,
    #                    form subdiagonalized [a] orbitals.
    #         include := include cut orbitlas into selfenergy embedding.
    #     '''
    #     h_mm = self.H
    #     s_mm = self.S
    #     nbf = h_mm.shape[-1]
    #
    #     c_MM, e_aj = subdiagonalize_atoms(calc, h_mm, s_mm, a)
    #     bfs_imp = get_bfs_indices(calc, a)
    #     # In [a] but not in active space
    #     bfs_not_m = [bfs_imp[i] for i,
    #                                 eigval in enumerate(flatten(e_aj))
    #                  if not key(eigval)] #Take complementary
    #     # Active space
    #     bfs_m = list(np.setdiff1d(bfs_imp,bfs_not_m))
    #     # o := orbitals other atoms not in [a]
    #     bfs_m_and_o_i = list(np.setdiff1d(range(nbf),bfs_not_m))
    #     nbfs_m_and_o = len(bfs_m_and_o_i)
    #
    #     #Apply subdiagonalization
    #     self.apply_rotation(c_MM)
    #
    #     if cutoff<np.inf:
    #         # h_imp = get_subspace(self.H, bfs_imp)
    #         # s_imp = get_subspace(self.S, bfs_imp)
    #         #Effective active space := activespace and effective embedding
    #         bfs_eff_i = []
    #         bfs_eff_imp = []
    #         for bfm in bfs_m: # for each bfm (bfm := basis of active)
    #             row_m_imp = abs(self.H[bfm,bfs_imp])
    #             #Index of couplings in [a]
    #             coupling  = np.where(row_m_imp>cutoff)[0]
    #             bfs_eff_imp = np.union1d(bfs_eff_imp, coupling)
    #             #Effective active embedding for bfm
    #             bfs_emb_i = np.arange(self.H.shape[0])[bfs_imp][coupling]
    #             #Effective activespace
    #             bfs_eff_i = np.union1d(bfs_eff_i, bfs_emb_i)
    #         #Unify with others
    #         bfs_eff_i = bfs_eff_i.tolist()
    #         bfs_m_and_o_i = np.union1d(bfs_eff_i, bfs_m_and_o_i).tolist()
    #         #Modify activespace with effective active space
    #         h_imp = get_subspace(self.H, bfs_eff_i)
    #         s_imp = get_subspace(self.S, bfs_eff_i)
    #     else:
    #         h_imp = get_subspace(self.H, bfs_m)
    #         s_imp = get_subspace(self.S, bfs_m)
    #
    #
    #     if include:
    #         from .internalselfenergy import InternalSelfEnergy
    #         #Subdiagonalized space of [a]
    #         h_mm = get_subspace(self.H, bfs_imp)
    #         s_mm = get_subspace(self.S, bfs_imp)
    #         #Indecixes in impurities [a]
    #         bfs_not_m_imp = [i for i,
    #                                eigval in enumerate(flatten(e_aj))
    #                          if not key(eigval)] #Take complementary
    #         # Active space
    #         bfs_m_imp = list(np.setdiff1d(np.arange(len(h_mm)),bfs_not_m_imp))
    #         hs_mm, hs_ii, hs_im =  extract_orthogonal_subspaces(h_mm,
    #                                                             s_mm,
    #                                                             bfs_m_imp, #Modify
    #                                                             bfs_not_m_imp, #Modify
    #                                                             orthogonal=orthogonal)
    #         #Embed rest of subdiagonalized orbitas (in [a]).
    #         h_im = np.zeros((len(bfs_not_m),len(self.H)), complex)
    #         s_im = np.zeros((len(bfs_not_m),len(self.H)), complex)
    #         h_im[:,bfs_m] = hs_im[0]
    #         s_im[:,bfs_m] = hs_im[1]
    #         # h_im.take(bfs_m_and_o_i, axis=1)
    #         # s_im.take(bfs_m_and_o_i, axis=1)
    #         selfenergy = InternalSelfEnergy(hs_ii, (h_im, s_im)) #Coupling to leads is None!
    #         self.selfenergies.append(selfenergy)
    #
    #
    #     #Take (effective) activespace
    #     self.take_bfs(bfs_m_and_o_i, apply=True)
    #
    #     return h_imp, s_imp

    def order_species(self, calc, sort_diag=False, apply=False):

        atoms = calc.atoms

        # Order atomic species
        symbols = np.array(atoms.symbols)
        species = set(symbols)
        indices_xa = [None for _ in range(len(species))]
        for e, elem in enumerate(species):
            indices_xa[e] = np.where(symbols==elem)[0]

        # Subdiagonalize atoms
        p_mm = self.subdiagonalize_atoms(calc, apply=False)
        # Order orbitals within species
        u_mm = self.order_subset(calc, indices_xa,
                                 sort_diag=sort_diag, apply=False)
        c_mm = p_mm.dot(u_mm)
        if apply:
            self.apply_rotation(c_mm)
            return

        return c_mm

    def order_subset(self, calc, indices_xa, sort_diag=False, apply=False):
        '''This function is supposed to be used in subdiagonalized Hamiltoninans to
        order the atomic orbital indices within groups [x] of atoms [a] specified by
        indices_xa. [a] can be a list of lists and/or lists and ints.'''
        h_mm = self.H
        s_mm = self.S
        m = h_mm.shape[-1]

        perm_list = []
        for indices_a in indices_xa:
            try:
                indices_a[0][0]
            except TypeError:
                # indices_a is a list of int
                bfs_i = get_bfs_indices(calc, indices_a, 'append')
            else:
                # indices_a is a list of lists of atoms forming a block
                bfs_i = []
                for jj in indices_a:
                    bfs_i.append(get_bfs_indices(calc, jj))
            bfs_i = np.array(bfs_i)
            if sort_diag:
                # Order diagonal in subset
                bfs_i = bfs_i.T
            perm_list.extend(bfs_i.flatten())
        # Rotation matrix
        c_mm = np.eye(m).take(perm_list, axis=1)
        if apply:
            self.apply_rotation(c_mm)
            return

        return c_mm


    def get_activespace(self, calc, a, key=lambda x: abs(x)<np.inf, orthogonal=True, inverse=False):
        '''
            key     := Apply condition to eigenvalues of subdiagonalized atoms.
            inverse := Get embedding instead. Put activespace in selfenergy
        '''
        from .internalselfenergy import InternalSelfEnergy

        h_mm = self.H
        s_mm = self.S
        nbf = h_mm.shape[-1]

        c_MM, e_aj = subdiagonalize_atoms(calc, h_mm, s_mm, a)
        bfs_imp = get_bfs_indices(calc, a)
        # Active space
        bfs_m = [bfs_imp[i] for i,
                                eigval in enumerate(flatten(e_aj))
                 if key(eigval)]
        # Embedding
        bfs_i = list(np.setdiff1d(range(nbf),bfs_m))
        nbf_i = len(bfs_i)

        hs_mm, hs_ii, hs_im =  extract_orthogonal_subspaces(h_mm,
                                                            s_mm,
                                                            bfs_m.copy(), # Do not modify
                                                            bfs_i.copy(), # Do not modify
                                                            c_MM,
                                                            orthogonal)

        # Reduce h_im, s_im
        for selfenergy in self.selfenergies:
            try:
                sigma_mm = np.empty((nbf_i,nbf_i), complex)
                h_im = selfenergy.h_im.take(bfs_i, axis=1)
                s_im = selfenergy.s_im.take(bfs_i, axis=1)
            except IndexError:
                print('selfenergies already ok!')
            else:
                selfenergy.h_im = h_im
                selfenergy.s_im = s_im
                selfenergy.sigma_mm = sigma_mm

        if inverse: #Take embedding instead. Put activespace in embedding
            #Activespace selfenergy
            hs_mi = tuple(hs_im[i].T.conj() for i in range(2))
            selfenergy = InternalSelfEnergy(hs_mm, hs_mi)

            if hasattr(self, 'Ginv'):
                self.Ginv = np.empty(hs_ii[0].shape, complex)

            self.__init__(hs_ii[0], hs_ii[1], self.selfenergies+[selfenergy])

            return e_aj

        #Embedding selfenergy
        selfenergy = InternalSelfEnergy(hs_ii, hs_im,
                                        selfenergies=self.selfenergies)

        if hasattr(self, 'Ginv'):
            self.Ginv = np.empty(hs_mm[0].shape, complex)

        self.__init__(hs_mm[0], hs_mm[1], [selfenergy])

        return e_aj


def set_pzd_carbons(calc, coupledhamiltonian, ext_C=None, int_C=None, Ce=None,
                    Ci=None, apply=True):
    '''This function takes the pz- and d- orbitals of a scattering region.
    coupledhamiltonian can be either a CoupledHamiltonian or GreenFunction.
    If apply is set to False, the indices of the pzd- orbitals are returned
    instead. Note that this function modufies the '''
    from qtpyt.analysis.tk_analysis import get_external_internal
    #
    if (ext_C is None) or (int_C is None):
        ext_C, int_C = get_external_internal(calc.atoms, 'C')

    if (Ce is None) or (Ci is None):
        Ce = [3,6,10,12]
        Ci = [3,6,10,11]

    bfs_ext_i = get_bfs_indices(calc, ext_C, method='append')
    bfs_int_i = get_bfs_indices(calc, int_C, method='append')

    #Take pz- and d- orbitas
    bfs_ext_pzd_i = np.array(bfs_ext_i)[:,Ce]
    bfs_int_pzd_i = np.array(bfs_int_i)[:,Ci]

    #Activespace pz- and d- orbitas
    bfs_m = np.union1d(bfs_int_pzd_i, bfs_ext_pzd_i)
    indices_C = np.union1d(ext_C, int_C)

    #Rotate scattering region
    coupledhamiltonian.subdiagonalize_atoms(calc, indices_C, apply=True)

    if apply:
        #Reduce matrix
        coupledhamiltonian.take_bfs(bfs_m, apply=True)
        return

    return bfs_m

def set_pz_d_embedding(calc, coupledhamiltonian, apply=True):
    ''''''
    from .internalselfenergy import InternalSelfEnergy

    h_mm = coupledhamiltonian.H
    s_mm = coupledhamiltonian.S
    nbf = h_mm.shape[-1]

    bfs_pzd = set_pzd_carbons(calc, coupledhamiltonian, apply=False)
    bfs_pz  = bfs_pzd[::4]
    bfs_d   = np.setdiff1d(bfs_pzd, bfs_pz)

    #Pzd- subspace
    h_pzd = get_subspace(h_mm, bfs_pzd)
    s_pzd = get_subspace(s_mm, bfs_pzd)

    #Pz- and d- indices in pzd- subspace
    bfs_pzd_pz  = np.arange(0,len(bfs_pzd),4)
    bfs_pzd_d   = np.setdiff1d(range(len(bfs_pzd)), bfs_pzd_pz)

    hs_mm, hs_ii, hs_im =  extract_orthogonal_subspaces(h_pzd,
                                                        s_pzd,
                                                        bfs_pzd_pz, #Modify
                                                        bfs_pzd_d, #Modify
                                                        orthogonal=True)

    h_im = np.zeros((len(bfs_d), nbf), dtype=complex)
    s_im = np.zeros((len(bfs_d), nbf), dtype=complex)
    h_im[:,bfs_pz] = hs_im[0]
    s_im[:,bfs_pz] = hs_im[1]
    selfenergy = InternalSelfEnergy(hs_ii, (h_im, s_im))

    if apply:
        coupledhamiltonian.take_bfs(bfs_pz, apply=True)
        coupledhamiltonian.selfenergies.append(selfenergy)
        return

    return bfs_pz, selfenergy


def set_pz_embed_rest(calc, coupledhamiltonian, apply=True):
    '''This function takes the pz- and d- orbitals of a scattering region.
    coupledhamiltonian can be either a CoupledHamiltonian or GreenFunction.
    If apply is set to False, the indices of the pzd- orbitals are returned
    instead. Note that this function modufies the '''
    from .internalselfenergy import InternalSelfEnergy

    h_mm = coupledhamiltonian.H
    s_mm = coupledhamiltonian.S
    nbf = h_mm.shape[-1]

    #
    indices_C = np.where(calc.atoms.symbols=='C')[0]

    #Rotate scattering region
    coupledhamiltonian.subdiagonalize_atoms(calc, indices_C, apply=True)

    #Carbon atom basis function indices
    bfs_C = np.array(get_bfs_indices(calc, indices_C, method='append'))
    #Pz-
    bfs_pz = bfs_C[:,3].flatten()
    #Rest-
    bfs_rest = np.setdiff1d(bfs_C, bfs_pz)

    #Carbon- subspace
    h_cc = get_subspace(h_mm, bfs_C.flatten())
    s_cc = get_subspace(s_mm, bfs_C.flatten())

    #Pz- and rest- indices in Carbon atoms subspace
    bfs_C_pz = np.arange(3, bfs_C.size, bfs_C.shape[1])
    bfs_C_rest  = np.setdiff1d(range(bfs_C.size), bfs_C_pz)

    hs_mm, hs_ii, hs_im =  extract_orthogonal_subspaces(h_cc,
                                                        s_cc,
                                                        bfs_C_pz, #Modify
                                                        bfs_C_rest, #Modify
                                                        orthogonal=True)

    h_im = np.zeros((len(bfs_C_rest), nbf), dtype=complex)
    s_im = np.zeros((len(bfs_C_rest), nbf), dtype=complex)
    h_im[:,bfs_pz] = hs_im[0]
    s_im[:,bfs_pz] = hs_im[1]
    selfenergy = InternalSelfEnergy(hs_ii, (h_im, s_im))

    if apply:
        coupledhamiltonian.take_bfs(bfs_pz, apply=True)
        coupledhamiltonian.selfenergies.append(selfenergy)
        return

    return bfs_pz, selfenergy



import matplotlib.pyplot as plt
import numpy as np
import scipy.linalg as linalg


def sort_eigs(alpha, betha, Z, h_l, h_r, flag):
    """
    The function provides classification and sorting of the eigenvalues
    and eigenvectors. The sorting procedure is described in
    [M. Wimmer, Quantum transport in nanostructures: From computational concepts to
    spintronics in graphene and magnetic tunnel junctions, 2009, ISBN-9783868450255]
    The algorithm puts left-propagating solutions into upper-left block and right-propagating
    solutions - into the lower-right block.
    The sort is performed by absolute value of the eigenvalues. For those whose
    absolute value equals one, classification is performed computing their group velocity
    using eigenvectors. The function also orthogonalizes the eigenvectors contained in the matrix Z
    that correspond to degenerate eigenvalues.
    :param alpha:
    :param betha:
    :param Z:
    :param h_l:
    :param h_r:
    :return:
    """

    margin = 1e-3
    angle_margin = 1e-4
    ans = []

    eigenvals = alpha / betha

    inds = np.where((np.abs(eigenvals) < 1.0 + margin) * (np.abs(eigenvals) > 1.0 - margin))[0]
    inds_deg = []

    for j, j1 in enumerate(inds[:-1]):
        inds_deg.append([j1])
        for j2 in inds[j + 1:]:
            if np.abs(np.real(eigenvals[j1] - eigenvals[j2])) < angle_margin and \
                    np.abs(np.imag(eigenvals[j1] - eigenvals[j2])) < angle_margin:
                inds_deg[j].append(j2)

    inds_deg = [i for i in inds_deg if len(i) > 1]

    for item in inds_deg:
        phi = np.matrix(Z[h_r.shape[0]:, item])
        oper = 1j * np.matrix(h_r * eigenvals[item[0]] - h_l * np.conj(eigenvals[item[0]]))
        _, vects = np.linalg.eig(phi.H * oper * phi)
        print(vects)
        Z[h_r.shape[0]:, item] = phi * np.matrix(vects)

    phi = np.matrix(Z)

    for j, item in enumerate(zip(alpha, betha)):

        if np.abs(item[1]) != 0.0:
            eigenval = item[0] / item[1]
        else:
            eigenval = 1e21

        if np.abs(eigenval) > 1.0 + margin:
            ans.append(False)
        elif np.abs(eigenval) < 1.0 - margin:
            ans.append(True)
        else:
            gv = np.imag(2 * phi[h_r.shape[0]:, j].H * np.matrix(h_r) * phi[h_r.shape[0]:, j])
            # gv = group_velocity(phi[h_r.shape[0]:, j], eigenval, h_r)

            if flag:
                gv = group_velocity(phi[h_r.shape[0]:, j], np.conj(eigenval), h_r)

            print("Group velocity is ", gv, np.angle(eigenval))
            if gv > 0:
                ans.append(False)
            else:
                ans.append(True)

    return ans


def iterate_gf(E, h_0, h_l, h_r, gf, num_iter):
    for j in range(num_iter):
        gf = h_r * np.linalg.pinv(E * np.identity(h_0.shape[0]) - h_0 - gf) * h_l

    return gf


def surface_greens_function_poles(E, h_l, h_0, h_r):
    """
    Computes eigenvalues and eigenvectors for the complex band structure problem.
    Here, the energy E is a parameter, and the eigenvalues correspond to wave vectors as `exp(ik)`.
    :param E:     energy
    :type E:      float
    :param h_l:   left block of three-block-diagonal Hamiltonian
    :param h_0:   central block of three-block-diagonal Hamiltonian
    :param h_r:   right block of three-block-diagonal Hamiltonian
    :return:      eigenvalues, k, and eigenvectors, U,
    :rtype:       numpy.matrix, numpy.matrix
    """

    main_matrix = np.block([[np.zeros(h_0.shape), np.identity(h_0.shape[0])],
                            [-h_l, E * np.identity(h_0.shape[0]) - h_0]])

    overlap_matrix = np.block([[np.identity(h_0.shape[0]), np.zeros(h_0.shape)],
                               [np.zeros(h_0.shape), h_r]])

    alpha, betha, _, eigenvects, _, _ = linalg.lapack.cggev(main_matrix, overlap_matrix)

    eigenvals = np.zeros(alpha.shape, dtype=np.complex128)

    for j, item in enumerate(zip(alpha, betha)):

        if np.abs(item[1]) != 0.0:
            eigenvals[j] = item[0] / item[1]
        else:
            eigenvals[j] = 1e10

    # sort absolute values
    ind = np.argsort(np.abs(eigenvals))
    eigenvals = eigenvals[ind]
    eigenvects = eigenvects[:, ind]

    vals = np.copy(eigenvals)
    roll = np.where((0.999 < np.abs(vals)) * (np.abs(vals) < 1.001))

    if np.size(roll) != 0:
        roll = h_0.shape[0] - int(round(np.mean(roll)))
        if roll != 0:
            print(roll)
            eigenvals = np.roll(eigenvals, roll)
            eigenvects = np.roll(eigenvects, roll, axis=1)
            eigenvals[:abs(roll)] = 0.0
            eigenvals[-abs(roll):] = 1e10
            vals = np.copy(eigenvals)

    mask1 = np.abs(vals) < 0.999
    mask2 = np.abs(vals) > 1.001
    vals = np.angle(vals)

    vals[mask1] = -5
    vals[mask2] = 5
    ind = np.argsort(vals, kind='mergesort')

    eigenvals = eigenvals[ind]
    eigenvects = eigenvects[:, ind]

    eigenvects = eigenvects[h_0.shape[0]:, :]
    eigenvals = np.matrix(np.diag(eigenvals))
    eigenvects = np.matrix(eigenvects)

    norms = linalg.norm(eigenvects, axis=0)
    norms = np.array([1e30 if np.abs(norm) < 0.000001 else norm for norm in norms])
    eigenvects = eigenvects / norms[np.newaxis, :]

    return eigenvals, eigenvects


def surface_greens_function_poles_Shur(E, h_l, h_0, h_r):
    """
    Computes eigenvalues and eigenvectors for the complex band structure problem.
    Here, the energy E is a parameter, and the eigenvalues correspond to wave vectors as `exp(ik)`.
    :param E:     energy
    :type E:      float
    :param h_l:   left block of three-block-diagonal Hamiltonian
    :param h_0:   central block of three-block-diagonal Hamiltonian
    :param h_r:   right block of three-block-diagonal Hamiltonian
    :return:      eigenvalues, k, and eigenvectors, U,
    :rtype:       numpy.matrix, numpy.matrix
    """

    main_matrix = np.block([[np.zeros(h_0.shape), np.identity(h_0.shape[0])],
                            [-h_l, E * np.identity(h_0.shape[0]) - h_0]]).astype(np.complex64)

    overlap_matrix = np.block([[np.identity(h_0.shape[0]), np.zeros(h_0.shape)],
                               [np.zeros(h_0.shape), h_r]]).astype(np.complex64)

    sort = lambda a, b, z: sort_eigs(a, b, z, h_l, h_r, False)

    AA, BB, alpha, betha, eigv_right, eigv_left = linalg.ordqz(main_matrix, overlap_matrix,
                                                               output='complex', sort=sort)

    main_matrix1 = np.block([[np.zeros(h_0.shape), np.identity(h_0.shape[0])],
                             [-h_r, E * np.identity(h_0.shape[0]) - h_0]]).astype(np.complex64)

    overlap_matrix1 = np.block([[np.identity(h_0.shape[0]), np.zeros(h_0.shape)],
                                [np.zeros(h_0.shape), h_l]]).astype(np.complex64)

    sort1 = lambda a, b, z: sort_eigs(a, b, z, h_r, h_l, True)

    AA1, BB1, alpha1, betha1, eigv_right1, eigv_left1 = linalg.ordqz(main_matrix1, overlap_matrix1,
                                                                     output='complex',
                                                                     sort=sort1)

    eigv_left = np.matrix(eigv_left)
    eigv_left1 = np.matrix(eigv_left1)

    return h_r * eigv_left[h_0.shape[0]:, :h_0.shape[0]] * np.linalg.pinv(eigv_left[:h_0.shape[0], :h_0.shape[0]]), \
           h_l * eigv_left1[h_0.shape[0]:, :h_0.shape[0]] * np.linalg.pinv(eigv_left1[:h_0.shape[0], :h_0.shape[0]])


def group_velocity(eigenvector, eigenvalue, h_r):
    """
    Computes the group velocity of wave packets from their wave vectors
    :param eigenvector:
    :param eigenvalue:
    :param h_r:
    :return:
    """

    return np.imag(eigenvector.H * h_r * eigenvalue * eigenvector)


def surface_greens_function(E, h_l, h_0, h_r):
    """
    Computes eigenvalues and eigenvectors for the complex band structure problem.
    Here energy E is a parameter, adn the eigenvalues are wave vectors.
    :param E:     energy
    :type E:      float
    :param h_l:   left block of three-block-diagonal Hamiltonian
    :param h_0:   central block of three-block-diagonal Hamiltonian
    :param h_r:   right block of three-block-diagonal Hamiltonian
    :return:      surface Green's function, for left and right sides
    """

    vals, vects = surface_greens_function_poles(E, h_l, h_0, h_r)
    vals = np.diag(vals)

    u_right = np.matrix(np.zeros(h_0.shape, dtype=np.complex))
    u_left = np.matrix(np.zeros(h_0.shape, dtype=np.complex))
    lambda_right = np.matrix(np.zeros(h_0.shape, dtype=np.complex))
    lambda_left = np.matrix(np.zeros(h_0.shape, dtype=np.complex))

    alpha = 0.001

    for j in range(h_0.shape[0]):
        if np.abs(vals[j]) > 1.0 + alpha:

            lambda_left[j, j] = vals[j]
            u_left[:, j] = vects[:, j]

            lambda_right[j, j] = vals[-j + 2 * h_0.shape[0] - 1]
            u_right[:, j] = vects[:, -j + 2 * h_0.shape[0] - 1]

        elif np.abs(vals[j]) < 1.0 - alpha:
            lambda_right[j, j] = vals[j]
            u_right[:, j] = vects[:, j]

            lambda_left[j, j] = vals[-j + 2 * h_0.shape[0] - 1]
            u_left[:, j] = vects[:, -j + 2 * h_0.shape[0] - 1]

        else:

            gv = group_velocity(vects[:, j], vals[j], h_r)
            # ind = np.argmin(np.abs(np.angle(vals[h_0.shape[0]:]) + np.angle(vals[j])))
            print("Group velocity is ", gv, np.angle(vals[j]))

            if gv > 0:

                lambda_left[j, j] = vals[j]
                u_left[:, j] = vects[:, j]

                lambda_right[j, j] = vals[-j + 2 * h_0.shape[0] - 1]
                u_right[:, j] = vects[:, -j + 2 * h_0.shape[0] - 1]

            else:
                lambda_right[j, j] = vals[j]
                u_right[:, j] = vects[:, j]

                lambda_left[j, j] = vals[-j + 2 * h_0.shape[0] - 1]
                u_left[:, j] = vects[:, -j + 2 * h_0.shape[0] - 1]

            # lambda_right[j, j] = vals[j]
            # u_right[:, j] = vects[:, j]
            #
            # lambda_left[j, j] = vals[-j + 2 * h_0.shape[0] - 1]
            # u_left[:, j] = vects[:, -j + 2 * h_0.shape[0] - 1]

    sgf_l = h_r * u_right * lambda_right * np.linalg.pinv(u_right)
    sgf_r = h_l * u_left * lambda_right * np.linalg.pinv(u_left)

    # sgf_l = u_right[h_0.shape[0]:, :] * np.linalg.pinv(u_right[:h_0.shape[0], :])
    # sgf_r = h_l * u_left * lambda_right * np.linalg.pinv(u_left)

    return iterate_gf(E, h_0, h_l, h_r, sgf_l, 2), iterate_gf(E, h_0, h_r, h_l, sgf_r, 2), \
           lambda_right, lambda_left, vals

    # return h_r * u_right * lambda_right * np.linalg.pinv(u_right), \
    #        h_l * u_left * lambda_right * np.linalg.pinv(u_left), \
    #        lambda_right, lambda_left, vals



from itertools import product, accumulate
from qtpyt import xp
import numpy as np
from qtpyt.abc._kernels import (dotdiag, dottrace)

def _boudscheck(i,j,N):
    if abs(i)>N:
        raise IndexError(f"Index {i} is out of bound for axis 0 with size {N}")
    if abs(j)>N:
        raise IndexError(f"Index {j} is out of bound for axis 1 with size {N}")
    if abs(j-i)>1:
        raise IndexError(f"Index {i,j} is out of tridiagonal")


def _wraparound(i,j,N):
    return i%N, j%N


class _BTBuffer:
    def __init__(self, N):
        self.N = N
        self.m_qii = [None for _ in range(N)]
        self.m_qij = [None for _ in range(N-1)]
        self.m_qji = [None for _ in range(N-1)]

    def __getitem__(self, pos):
        i, j = pos
        _boudscheck(i,j,self.N)
        i, j =_wraparound(i,j,self.N)
        if i==j:
            return self.m_qii[i]
        if i>j:
            return self.m_qji[i-1]
        return self.m_qji[i-1]

    def __setitem__(self, pos, val):
        i, j = pos
        _boudscheck(i,j,self.N)
        i, j =_wraparound(i,j,self.N)
        if i==j:
            self.m_qii[i] = val 
        if i>j:
            self.m_qji[i-1] = val
        self.m_qji[i-1] = val

    def __iter__(self):
        yield self.m_qii[0]
        for q in range(1,self.N):
            yield self.m_qij[q-1]
            yield self.m_qji[q-1]
            yield self.m_qii[q]

    def __len__(self):
        return self.N


def empty_buffer(N):
    return _BTBuffer(N)


class BTMatrix(_BTBuffer):
    """Helper class for Block Tridiagonal matrices.
    
    Args:
        m_qii : (list(xp.ndarray), len=N)
            List of block matrices on the diagonal. 

        m_qij : (list(xp.ndarray), len=N-1)
            List of block matrices on the upper diagonal. 

        m_qji : (list(xp.ndarray), len=N-1)
            List of block matrices on the lower diagonal. 
    """
    def __init__(self, m_qii, m_qij, m_qji):
        super().__init__(len(m_qii))
        self.m_qii = m_qii
        self.m_qij = m_qij
        self.m_qji = m_qji
        
        shapes = [m.shape[0] for m in self.m_qii]
        self.m = sum(shapes)
        self.shape = np.fromiter(product(shapes, shapes), 'i8,i8').reshape(self.N,self.N)
        self._nodes = np.fromiter(accumulate([0]+shapes),'i8')

    def todense(self):
        order = 'C' if self.m_qii[0].flags.c_contiguous else 'F'
        M = xp.zeros(2*(self.m,), dtype=self.m_qii[0].dtype, order=order)
        M[self.ix_(0,0)] = self.m_qii[0]
        for q in range(1,self.N):
            M[self.ix_(q-1,q)] = self.m_qij[q-1]
            M[self.ix_(q,q-1)] = self.m_qji[q-1]
            M[self.ix_(q,q)] = self.m_qii[q]
        return M

    def ix_(self,i,j=None):
        if j is None: j=i
        il = self._nodes[i]
        jl = self._nodes[j]
        iu = il + self.shape[i,j][0]
        ju = jl + self.shape[i,j][1]
        return (slice(il,iu),
                slice(jl,ju))

    def __eq__(self,other,rtol=1e-05, atol=1e-08):
        allclose = lambda x, y: xp.testing.assert_allclose(x,y,rtol=rtol, atol=atol)
        for a, b in zip(self, other):
            allclose(a, b)

    def diagonal(self):
        return xp.concatenate([self.m_qii[q].diagonal() for q in range(self.N)])

    def trace(self):
        return xp.sum([self.m_qii[q].trace() for q in range(self.N)])

    def dot(self, other):
        out = empty_buffer(self.N)
        out.m_qii[0] = xp.dot(self.m_qii[0], other.m_qii[0])
        for q in range(self.N-1):
            out.m_qii[q+1] = xp.dot(self.m_qii[q+1], other.m_qii[q+1])
            out.m_qii[q] += xp.dot(self.m_qij[q], other.m_qji[q])
            out.m_qii[q+1] += xp.dot(self.m_qji[q], other.m_qij[q])
        return BTMatrix(out.m_qii, out.m_qij, out.m_qji)

    def dotdiag(self, other):
        out = [xp.zeros(self.m_qii[q].shape[0],self.m_qii[q].dtype) for q in range(self.N)]
        dotdiag(self.m_qii[0], other.m_qii[0], out[0])
        for q in range(self.N-1):
            dotdiag(self.m_qii[q+1], other.m_qii[q+1], out[q+1])
            dotdiag(self.m_qij[q], other.m_qji[q], out[q])
            dotdiag(self.m_qji[q], other.m_qij[q], out[q+1])
        return xp.concatenate(out)

    def dottrace(self, other):
        out = complex(0.)
        out = dottrace(self.m_qii[0], other.m_qii[0], out)
        for q in range(self.N-1):
            out = dottrace(self.m_qii[q+1], other.m_qii[q+1], out)
            out = dottrace(self.m_qij[q], other.m_qji[q], out)
            out = dottrace(self.m_qji[q], other.m_qij[q], out)
        return out

    _specials = ['__add__','__sub__','__iadd__',
                 '__isub__','__mul__','__imul__',
                 '__radd__','__rsub__','__rmul__']

    @classmethod
    def _create(cls):
        """
        TODO :: Add option to add diagonal array or constant.
        """
        ##https://code.activestate.com/recipes/496741-object-proxying/
        
        def make_method(name):
            def method(self, other, **kwargs):
                op = getattr(xp.ndarray, name)
                out = empty_buffer(self.N)
                if isinstance(other, BTMatrix):
                    out.m_qii[0] = op(self.m_qii[0], other.m_qii[0])
                    for q in range(1,self.N):
                        out.m_qii[q] = op(self.m_qii[q], other.m_qii[q])
                        out.m_qij[q-1] = op(self.m_qij[q-1], other.m_qij[q-1])
                        out.m_qji[q-1] = op(self.m_qji[q-1], other.m_qji[q-1])
                else:
                    out.m_qii[0] = op(self.m_qii[0], other)
                    for q in range(1,self.N):
                        out.m_qii[q] = op(self.m_qii[q], other)
                        out.m_qij[q-1] = op(self.m_qij[q-1], other)
                        out.m_qji[q-1] = op(self.m_qji[q-1], other)
                return BTMatrix(out.m_qii, out.m_qij, out.m_qji)
            return method
        
        namespace = {}
        for name in cls._specials:
            namespace[name] = make_method(name)
        return type("%s(%s)" % (cls.__name__, xp.ndarray.__name__), (cls,), namespace)

    def __new__(cls, *args, **kwargs):
        theclass = cls._create()
        ins = object.__new__(theclass)
        theclass.__init__(ins, *args, **kwargs)
        return ins



import numpy as np
import matplotlib.pyplot as plt
from scipy.linalg import eig, eigh
from scipy.constants import e, k

k_B = k / e # Boltzmann constant [eV/K] 8.6173303e-05

def fd(energy, ef, temp):
    return 1.0 / (1.0 + np.exp((energy - ef) / (k_B * temp)))


def approximant(energy, poles, residues):

    arg = np.array(energy)
    ans = np.zeros(arg.shape) + 0.5

    for j in range(len(poles)):
        if poles[j] > 0:
            ans = ans - residues[j] / (arg - 1j / poles[j]) - residues[j] / (arg + 1j / poles[j])

    return ans


def approximant_diff(energy, poles, residues):

    arg = np.array(energy)
    ans = np.zeros(arg.shape) + 0.5 * 0

    for j in range(len(poles)):
        if poles[j] > 0:
            ans = ans - residues[j] / (arg - 1j / poles[j]) ** 2 - residues[j] / (arg + 1j / poles[j]) ** 2

    return ans


def poles_and_residues(cutoff=2):

    b_mat = [1 / (2.0 * np.sqrt((2*(j+1) - 1)*(2*(j+1) + 1))) for j in range(0, cutoff-1)]
    b_mat = np.matrix(np.diag(b_mat, -1)) + np.matrix(np.diag(b_mat, 1))

    poles, residues = eig(b_mat)

    residues = np.array(np.matrix(residues))
    # arg = np.argmax(np.abs(residues), axis=0)

    residues = 0.25 * np.array([np.abs(residues[0, j])**2 / (poles[j] ** 2) for j in range(residues.shape[0])])

    return poles, residues


def zero_fermi(nzp):
    '''Compute poles (zp) and residues (Rp) of fermi function.'''
    N = nzp
    M = 2*N

    A = np.zeros((2*M,2*M))
    B = np.zeros((2*M,2*M))

    zp = np.zeros(2+M)
    Rp = np.zeros(2+M)

    for i in range(1,M+1):
        B[i,i] = (2*i-1)

    for i in range(1,M):
        A[i,i+1] = -0.5;
        A[i+1,i] = -0.5;

    a = np.zeros(M*M)
    b = np.zeros(M*M)

    for i in range(M):
        for j in range(M):
            a[j*M+i] = A[i+1,j+1]
            b[j*M+i] = B[i+1,j+1]

    a.shape = (M,M)
    b.shape = (M,M)

    eigvas, eigvecs = eigh(a,b)

    zp[:M] = eigvas

    for i in range(M,0,-1):
        zp[i] = zp[i-1]

    for i in range(1,M+1):
        zp[i] = 1.0/zp[i];

    a = eigvecs.T.flatten()

    for i in range(0,M):
        Rp[i+1] = -a[i*M]*a[i*M]*zp[i+1]*zp[i+1]*0.250;

    zp = -zp[1:N+1]
    Rp = Rp[1:N+1]

    return zp, Rp

def integrate_pdos(G, zp, Rp, mu=0, T=300):#nzp=100, R=1e10):

    # zp, Rp = zero_fermi(nzp)
    N = len(zp)#nzp

    beta = 1/(k_B*T)
    a_p = mu + 1j*zp/beta

    eta = G.eta
    G.eta = complex(0.)

    R = 1e10
    mu_0 = 1j*R*G.apply_retarded(1j*R, G.S).diagonal()

    mu_1 = np.zeros(len(G.H), complex)
    for i in range(N):
        mu_1 += G.apply_retarded(a_p[i], G.S).diagonal() * Rp[i]
    mu_1 *= -1j*4/beta

    rho = np.real(mu_0) + np.imag(mu_1)

    G.eta = eta

    return rho

if __name__=='__main__':

    a1, b1 = poles_and_residues(cutoff=2)
    a2, b2 = poles_and_residues(cutoff=10)
    a3, b3 = poles_and_residues(cutoff=30)
    a4, b4 = poles_and_residues(cutoff=50)
    a5, b5 = poles_and_residues(cutoff=100)

    energy = np.linspace(-5.7, 5.7, 3000)

    temp = 100
    fd0 = fd(energy, 0, temp)

    k_B = 8.61733e-5  # Boltzmann constant in eV
    energy = energy / (k_B * temp)

    ans1 = approximant(energy, a1, b1)
    ans2 = approximant(energy, a2, b2)
    ans3 = approximant(energy, a3, b3)
    ans4 = approximant(energy, a4, b4)
    ans5 = approximant(energy, a5, b5)

    plt.plot(fd0)
    plt.plot(ans1)
    plt.plot(ans2)
    plt.plot(ans3)
    plt.plot(ans4)
    plt.plot(ans5)
    plt.show()



import numpy as np
from functools import lru_cache
from .internalselfenergy import InternalSelfEnergy
from .tools import rotate_matrix, get_subspace
from scipy import linalg as la
# from gpaw.utilities.blas import gemm
# import _cppmodule as _cpp

# class LeadSelfEnergy(_cpp.LeadSelfEnergy, InternalSelfEnergy):
class LeadSelfEnergy(InternalSelfEnergy):
    conv = 1e-8 # Convergence criteria for surface Green function

    def __init__(self, hs_dii, hs_dij, hs_dim, eta=1e-4):
        self.h_ij, self.s_ij = hs_dij # coupling between principal layers
        # self.bias = 0
        InternalSelfEnergy.__init__(self, hs_dii, hs_dim, eta=eta)

    def set_bias(self, bias):
        self.bias = bias

    def get_lambda(self, energy):
        """Return the lambda (aka Gamma) defined by i(S-S^d).

        Here S is the retarded selfenergy, and d denotes the hermitian
        conjugate.
        """
        sigma_mm = self.retarded(energy)
        return 1.j * (sigma_mm - sigma_mm.T.conj())

    # # @lru_cache(maxsize=None)
    def get_Ginv(self, energy):
        # """The inverse of the retarded surface Green function"""
        z = energy - self.bias + self.eta * 1.j

        v_00 = z * self.s_ii.T.conj() - self.h_ii.T.conj()
        v_11 = v_00.copy()
        v_10 = z * self.s_ij - self.h_ij
        v_01 = z * self.s_ij.T.conj() - self.h_ij.T.conj()

        delta = self.conv + 1
        while delta > self.conv:
            lu, piv = la.lu_factor(v_11)
            a = la.lu_solve((lu, piv), v_01)
            b = la.lu_solve((lu, piv), v_10)
            v_01_dot_b = np.dot(v_01, b)
            v_00 -= v_01_dot_b
            v_11 -= np.dot(v_10, a)
            v_11 -= v_01_dot_b
            v_01 = -np.dot(v_01, a)
            v_10 = -np.dot(v_10, b)
            delta = abs(v_01).max()

        return v_00

    def apply_rotation(self, c_mm):
        # self.get_sgfinv.cache_clear()
        self.h_ii = self.h_ii.copy() # needed because self[0].h_ii is self[1].h_ii
        self.s_ii = self.s_ii.copy() # as self[0].h_ii -> H1[:nprinc,:nprinc]
        InternalSelfEnergy.apply_rotation(self, c_mm)
        self.h_ij = rotate_matrix(self.h_ij, c_mm)
        self.s_ij = rotate_matrix(self.s_ij, c_mm)

    def cutcoupling_bfs(self, bfs, apply=False):
        # self.get_sgfinv.cache_clear()
        h_pp, s_pp = InternalSelfEnergy.cutcoupling_bfs(self, bfs, apply)
        if apply:
            for m in bfs:
                self.h_ij[m, :] = 0.0
                self.s_ij[m, :] = 0.0
        return h_pp, s_pp

    def take_bfs(self, bfs, apply=False):
        # self.get_sgfinv.cache_clear()
        h_pp, s_pp, c_mm = InternalSelfEnergy.take_bfs(self, bfs, apply)
        if apply:
            self.h_ij = rotate_matrix(self.h_ij, c_mm) #np.dot(c_mm.T.conj(), self.h_ij)
            self.s_ij = rotate_matrix(self.s_ij, c_mm) #np.dot(c_mm.T.conj(), self.s_ij)
        return h_pp, s_pp


class BoxProbe:
    """Box shaped Buttinger probe.

    Kramers-kroning: real = H(imag); imag = -H(real)
    """
    def __init__(self, eta, a, b, energies, S, T=0.3):
        from qtpyt.Hilbert import hilbert
        se = np.empty(len(energies), complex)
        se.imag = .5 * (np.tanh(.5 * (energies - a) / T) -
                        np.tanh(.5 * (energies - b) / T))
        se.real = hilbert(se.imag)
        se.imag -= 1
        self.selfenergy_e = eta * se
        self.energies = energies
        self.S = S

    def retarded(self, energy):
        return self.selfenergy_e[self.energies.searchsorted(energy)] * self.S
